"""
This module contains tests for:

- test_kustomizations: Test if all kustomizations are ready
- test_helmreleases: Test if all the helmreleases are ready
- test_deployments: Test if all deployments are ready

Documentation for the kubernetes client:
* https://github.com/kubernetes-client/python/blob/master/kubernetes/README.md
* https://github.com/kubernetes-client/python/tree/master/examples
"""
import os
from kubernetes import client, config
from kubernetes.client.rest import ApiException

import pytest

# Helper functions

def check_resource(api, api_group, api_version, plural, name, namespace):
    """Returns status contition for resource in given namespace

    :param api: The api object used to query the custom objects
    :param api_group: The api group used to query the custom objects
    :type api_group: str
    :param api_version: The api version used to query the custom objects
    :type api_version: str
    :param plural: The custom resource's plural name
    :type plural: str
    :param name: The custom object's name
    :type name: str
    :param namespace: The custom object's namespace
    :type namespace: str
    :returns: The custom object's status
    :rtype: str
    """
    if not namespace:
        namespace = 'flux-system'
    print(f'Resource "{name}" in namespace "{namespace}" '
          f'(api version: {api_group}/{api_version}): ', end='')
    try:
        resource = api.get_namespaced_custom_object(
            group=api_group,
            version=api_version,
            plural=plural,
            name=name,
            namespace=namespace
        )
        status = 'False'
        for condition in resource['status']['conditions']:
            if condition['type'] == 'Ready':
                status = condition['status']
        print(status)
    except ApiException as ex:
        if ex.status == 404:
            status = 'Not found'
        else:
            raise
        print(f"**** NOT ready, status: {status} *****")
    except KeyError:
        status = 'Key error'
        print("resource key error: ")
        print(resource) # type: ignore (Ignore pyright reportUnboundVariable)
    return status

def check_custom_objects(api, api_group, api_version, plural, name, namespace):
    """
    Checks if all resoures of a given custom object are ready.

    We can't get a list of custom objects from all namespaces,
    so we have to iterate over all namespaces. See
    https://github.com/kubernetes-client/python/issues/1377

    :param api: The api object used to query the custom objects
    :param api_group: The api group used to query the custom objects
    :type api_group: str
    :param api_version: The api version used to query the custom objects
    :type api_version: str
    :param plural: The custom resource's plural name
    :type plural: str
    :param name: The custom object's name
    :type name: str
    :param namespace: The custom object's namespace
    :type namespace: str
    :returns: The custom object's status
    :rtype: str
    """
    failed = 0
    failed_resources = []

    if not namespace:
        namespace = 'flux-system'
    if name != 'all':
        status = check_resource(
            api=api,
            api_group=api_group,
            api_version=api_version,
            plural=plural,
            name=name,
            namespace=namespace)
        if status != 'True':
            failed += 1
            failed_resources.append(name)
    else:
        core_api = client.CoreV1Api()
        namespaces = core_api.list_namespace(watch=False)
        for namespace_object in namespaces.items:
            namespace_name = namespace_object.metadata.name
            custom_objects = api.list_namespaced_custom_object(
                group=api_group,
                version=api_version,
                plural=plural,
                namespace=namespace_name,
                watch=False
            )
            for custom_object in custom_objects['items']:
                resource_name = custom_object['metadata']['name']
                status = check_resource(
                    api=api,
                    api_group=api_group,
                    api_version=api_version,
                    plural=plural,
                    name=resource_name,
                    namespace=namespace_name)
                if status != 'True':
                    failed += 1
                    failed_resources.append(resource_name)
    assert failed == 0, f"Error: {failed} resources not ready ({failed_resources})!"


@pytest.fixture(autouse=True)
def run_around_tests():
    """
    Prepare kube config before running a test
    """
    cluster_dir = os.environ.get("CLUSTER_DIR")
    kubeconfig = os.path.join('..', str(cluster_dir), 'kube_config_cluster.yml')
    config.load_kube_config(config_file=kubeconfig)
    yield


# Pytest functions

@pytest.mark.resource
@pytest.mark.kustomizations
def test_kustomizations(resource, namespace):
    """
    Test if a single or all kustomization(s) installed by weave flux are in
    'Ready' state.

    If you want to test a single kustomization please provide both name and
    namespace, otherwise all kustomizations from all namespaces are tested.

    :param name: The resource object's name
    :type name: str
    :param namespace: The resource object's namespace
    :type namespace: str
    """

    print('\n')
    custom_objects = client.CustomObjectsApi()
    check_custom_objects(
        api=custom_objects,
        api_group='kustomize.toolkit.fluxcd.io',
        api_version='v1beta1',
        plural='kustomizations',
        name=resource,
        namespace=namespace)
